# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Models to fit the low-frequency part of the spectrum."""

import attrs
import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import brentq
from scipy.special import polygamma

__all__ = ("ExpPolyModel", "LorentzModel", "PadeModel", "SpectrumModel", "guess")


@attrs.define
class SpectrumModel:
    """Abstract base class for spectrum models.

    Subclasses must override all methods that raise ``NotImplementedError``.

    The first parameter must have a property that is used when constructing an initial guess:
    When the first parameter increases, the model should increase everywhere,
    and must allow for an arbitrary increase of the spectrum at all points.
    This is used to repair initial guesses that result in a partially negative spectrum.
    """

    scales: dict[str, float] = attrs.field(factory=dict, init=False)
    """A dictionary with essential scale information for the parameters and the cost function."""

    @property
    def name(self):
        raise NotImplementedError

    def bounds(self) -> list[tuple[float, float]]:
        """Return parameter bounds for the optimizer."""
        raise NotImplementedError

    @property
    def npar(self):
        """Return the number of parameters."""
        raise NotImplementedError

    def valid(self, pars: NDArray[float]) -> bool:
        """Return ``True`` when the parameters are within the feasible region."""
        result = np.ones(pars.shape[:-1], dtype=bool)
        for i, (pmin, pmax) in enumerate(self.bounds()):
            result &= pmin < pars[..., i]
            result &= pars[..., i] < pmax
        return result

    def which_invalid(self, pars) -> NDArray[bool]:
        """Return a boolean mask for the parameters outside the feasible region."""
        return np.array(
            [
                pmin >= par or par >= pmax
                for (pmin, pmax), par in zip(self.bounds(), pars, strict=True)
            ]
        )

    def configure_scales(
        self, timestep: float, freqs: NDArray[float], amplitudes: NDArray[float]
    ) -> NDArray[float]:
        """Store essential scale information in the ``scales`` attribute.

        Other methods may access this information,
        so this method should be called before performing any computations.
        """
        self.scales = {
            "freq_small": freqs[1],
            "freq_scale": freqs[-1],
            "timestep": timestep,
            "time_scale": 1 / freqs[-1],
            "amp_scale": np.median(abs(amplitudes[amplitudes != 0])),
        }

    @property
    def par_scales() -> NDArray[float]:
        """Return the scales of the parameters and the cost function."""
        raise NotImplementedError

    def get_par_nonlinear(self) -> NDArray[bool]:
        """Return a boolean mask for the nonlinear parameters.

        The returned parameters cannot be solved with the solve_linear method.
        Models are free to decide which parameters can be solved with linear regression.
        For example, some non-linear parameters may be solved with a linear regression
        after rewriting the regression problem in a different form.
        """
        raise NotImplementedError

    def sample_nonlinear_pars(
        self,
        rng: np.random.Generator,
        budget: int,
    ) -> NDArray[float]:
        """Return samples of the nonlinear parameters.

        Parameters
        ----------
        rng
            The random number generator.
        budget
            The number of samples to generate.
        freqs
            The frequencies for which the model spectrum amplitudes are computed.
        par_scales
            The scales of the parameters and the cost function.

        Returns
        -------
        samples
            The samples of the nonlinear parameters, array with shape ``(budget, num_nonlinear)``,
            where ``num_nonlinear`` is the number of nonlinear parameters.
        """
        raise NotImplementedError

    def solve_linear(
        self,
        freqs: NDArray[float],
        ndofs: NDArray[float],
        amplitudes: NDArray[float],
        weights: NDArray[float],
        nonlinear_pars: NDArray[float],
    ) -> NDArray[float]:
        """Use linear linear regression to solve a subset of the parameters.

        The default implementation in the base class assumes that the linear parameters
        are genuinely linear without rewriting the regression problem in a different form.

        Parameters
        ----------
        freqs
            The frequencies for which the model spectrum amplitudes are computed.
        amplitudes
            The amplitudes of the spectrum.
        ndofs
            The number of degrees of freedom at each frequency.
        weights
            Fitting weights, in range [0, 1], to use for each grid point.
        nonlinear_pars
            The values of the nonlinear parameters for which the basis functions are computed.

        Returns
        -------
        linear_pars
            The solved linear parameters.
        amplitudes_model
            The model amplitudes computed with the solved parameters.
        """
        nonlinear_pars = np.asarray(nonlinear_pars)
        if nonlinear_pars.ndim != 1:
            raise ValueError("The nonlinear parameters must be a 1D array.")
        nonlinear_mask = self.get_par_nonlinear()
        pars = np.ones(self.npar)
        pars[nonlinear_mask] = nonlinear_pars
        basis = self.compute(freqs, pars, deriv=1)[1][~nonlinear_mask]
        amplitudes_std = amplitudes / np.sqrt(0.5 * ndofs)
        rescaling = np.sqrt(weights) / amplitudes_std
        linear_pars = np.linalg.lstsq(
            (basis * rescaling).T,
            amplitudes * rescaling,
            # For compatibility with numpy < 2.0
            rcond=-1,
        )[0]
        amplitudes_model = np.dot(linear_pars, basis)
        return linear_pars, amplitudes_model

    def compute(
        self, freqs: NDArray[float], pars: NDArray[float], *, deriv: int = 0
    ) -> list[NDArray[float]]:
        """Compute the amplitudes of the spectrum model.

        Parameters
        ----------
        freqs
            The frequencies for which the model spectrum amplitudes are computed.
        pars
            The parameter vector.
            For vectorized calculations, the last axis corresponds to the parameter index.
        deriv
            The maximum order of derivatives to compute: 0, 1 or 2.

        Returns
        -------
        results
            A results list, index corresponds to order of derivative.
            The shape of the arrays in the results list is as follows:

            - For ``deriv=0``, the shape is ``(*vec_shape, len(freqs))``.
            - For ``deriv=1``, the shape is ``(*vec_shape, len(pars), len(freqs))``.
            - For ``deriv=2``, the shape is ``(*vec_shape, len(pars), len(pars), len(freqs))``

            If some derivatives are independent of the parameters,
            broadcasting rules may be used to reduce the memory footprint.
            This means that ``vec_shape`` may be replaced by a tuple of ones with the same length.
        """
        raise NotImplementedError

    def neglog_prior(self, pars: NDArray[float], *, deriv: int = 0) -> list[NDArray[float]]:
        """Minus logarithm of the prior probability density function, if any.

        Subclasses may implement (a very weak) prior, if any.
        """
        vec_shape = pars.shape[:-1]
        par_shape = pars.shape[-1:]
        result = [np.zeros(vec_shape)]
        if deriv > 0:
            result.append(np.zeros(vec_shape + par_shape))
        if deriv > 1:
            result.append(np.zeros(vec_shape + par_shape + par_shape))
        if deriv > 2:
            raise ValueError("Third or higher derivatives are not supported.")
        return result

    def derive_props(self, props: dict[str, NDArray[float]]):
        """Add the autocorrelation integral (and other properties) derived from the parameters.

        Parameters
        ----------
        props
            The properties dictionary, including the parameters and their uncertainties.
            Subclasses may add additional properties to this dictionary.
        """


def _convert_degrees(value: ArrayLike) -> NDArray[int]:
    """Convert the input to an array of integers."""
    value = np.asarray(value, dtype=int)
    if value.ndim == 0:
        return value[None]
    if value.ndim > 1:
        raise ValueError("Input must be a scalar or a 1D array.")
    value.sort()
    return value


def _validate_degrees(obj, attribute, value: NDArray[int]) -> None:
    """Validate the degrees of the polynomial."""
    if not np.all(value >= 0):
        raise ValueError("All degrees must be non-negative.")
    if len(value) == 0:
        raise ValueError("The list of degrees must not be empty.")
    if len(value) != len(set(value)):
        raise ValueError("The list of degrees must not contain duplicates.")


@attrs.define
class ExpPolyModel(SpectrumModel):
    """Exponential function of a linear combination of simple monomials."""

    degrees: NDArray[int] = attrs.field(converter=_convert_degrees, validator=_validate_degrees)
    """The degree of the monomials."""

    @property
    def name(self):
        degrees_str = ", ".join(map(str, self.degrees))
        return f"exppoly({degrees_str})"

    def bounds(self) -> list[tuple[float, float]]:
        """Return parameter bounds for the optimizer."""
        return [(-np.inf, np.inf)] * self.npar

    @property
    def npar(self):
        """Return the number of parameters."""
        return len(self.degrees)

    @property
    def par_scales(self) -> NDArray[float]:
        """Return the scales of the parameters and the cost function."""
        return self.scales["freq_scale"] ** -self.degrees

    def get_par_nonlinear(self) -> NDArray[bool]:
        """Return a boolean mask for the nonlinear parameters."""
        return np.zeros(self.npar, dtype=bool)

    def solve_linear(
        self,
        freqs: NDArray[float],
        ndofs: NDArray[float],
        amplitudes: NDArray[float],
        weights: NDArray[float],
        nonlinear_pars: NDArray[float],
    ) -> NDArray[float]:
        """Use linear linear regression to solve a subset of the parameters.

        This is a specialized implementation that rewrites the problem
        in a different form to solve all parameters with a linear regression.
        """
        if len(nonlinear_pars) != 0:
            raise ValueError("The number of nonlinear parameters must be exactly 0.")
        log_amplitudes = np.log(amplitudes)
        log_amplitudes_std = polygamma(1, 0.5 * ndofs)
        rescaling = np.sqrt(weights) / log_amplitudes_std
        expected_values = log_amplitudes * rescaling
        design_matrix = np.power.outer(freqs, self.degrees) * rescaling[:, None]
        pars = np.linalg.lstsq(design_matrix, expected_values, rcond=-1)[0]
        amplitudes_model = np.exp(np.dot(design_matrix, pars))
        return pars, amplitudes_model

    def compute(
        self, freqs: NDArray[float], pars: NDArray[float], *, deriv: int = 0
    ) -> list[NDArray[float]]:
        """See :py:meth:`SpectrumModel.compute`."""
        if not isinstance(deriv, int):
            raise TypeError("Argument deriv must be integer.")
        if deriv < 0:
            raise ValueError("Argument deriv must be zero or positive.")
        if freqs.ndim != 1:
            raise TypeError("Argument freqs must be a 1D array.")

        # Construct a basis of simple monomials.
        basis = np.power.outer(freqs, self.degrees).T

        # Compute model amplitudes and derivatives.
        func = np.exp(np.einsum("...p,pf->...f", pars, basis))
        results = [func]
        if deriv >= 1:
            results.append(np.einsum("...f,pf->...pf", func, basis))
        if deriv >= 2:
            results.append(np.einsum("...f,pf,qf->...pqf", func, basis, basis))
        if deriv >= 3:
            raise ValueError("Third or higher derivatives are not supported.")
        return results

    def derive_props(self, props: dict[str, NDArray[float]]):
        """Add the autocorrelation integral (and other properties) derived from the parameters."""
        # The logarithm of the autocorrelation integral is the first parameter,
        # which is assumed to be normally distributed.
        log_acint = props["pars"][0]
        log_acint_var = props["pars_covar"][0, 0]
        # The autocorrelation integral is the exponential of the first parameter,
        # and is therefore log-normally distributed.
        acint = np.exp(log_acint + 0.5 * log_acint_var)
        acint_var = (np.exp(log_acint_var) - 1) * np.exp(2 * log_acint + log_acint_var)
        acint_props = {
            "log_acint": log_acint,
            "log_acint_var": log_acint_var,
            "acint": acint,
            "acint_var": acint_var,
        }
        props.update(acint_props)


@attrs.define
class PadeModel(SpectrumModel):
    """A rational function model for the spectrum, a.k.a. a Padé approximation."""

    numer_degrees: NDArray[int] = attrs.field(
        converter=_convert_degrees, validator=_validate_degrees
    )
    """The degrees of the monomials in the numerator."""

    denom_degrees: NDArray[int] = attrs.field(
        converter=_convert_degrees, validator=_validate_degrees
    )
    """The degrees of the monomials in the denominator.

    Note that the leading term is always 1, and there is no need to include
    degree zero.
    """

    @property
    def name(self):
        numer_str = ", ".join(map(str, self.numer_degrees))
        denom_str = ", ".join(map(str, self.denom_degrees))
        return f"pade({numer_str}; {denom_str})"

    def bounds(self) -> list[tuple[float, float]]:
        """Return parameter bounds for the optimizer."""
        return [(-np.inf, np.inf)] * len(self.numer_degrees) + [(0, np.inf)] * len(
            self.denom_degrees
        )

    @property
    def npar(self):
        """Return the number of parameters."""
        return len(self.numer_degrees) + len(self.denom_degrees)

    @property
    def par_scales(self) -> NDArray[float]:
        """Return the scales of the parameters and the cost function."""
        return np.concatenate(
            [
                self.scales["amp_scale"]
                * self.scales["freq_scale"] ** (-np.array(self.numer_degrees)),
                self.scales["freq_scale"] ** (-np.array(self.denom_degrees)),
            ]
        )

    def get_par_nonlinear(self) -> NDArray[bool]:
        """Return a boolean mask for the nonlinear parameters."""
        return np.zeros(self.npar, dtype=bool)

    def solve_linear(
        self,
        freqs: NDArray[float],
        ndofs: NDArray[float],
        amplitudes: NDArray[float],
        weights: NDArray[float],
        nonlinear_pars: NDArray[float],
    ) -> NDArray[float]:
        """Use linear linear regression to solve a subset of the parameters.

        This is a specialized implementation that rewrites the problem
        in a different form to solve all parameters with a linear regression.
        """
        if len(nonlinear_pars) != 0:
            raise ValueError("The number of nonlinear parameters must be exactly 0.")
        # Rescale frequencies as a simple form of preconditioning.
        x = freqs / freqs[-1]

        # Construct bases of monomials.
        basis_n = np.power.outer(x, self.numer_degrees).T
        basis_d = np.power.outer(x, self.denom_degrees).T

        # Set up and solve linear regression problem.
        amplitudes_std = amplitudes / np.sqrt(0.5 * ndofs)
        rescaling = np.sqrt(weights) / amplitudes_std
        expected_values = amplitudes * rescaling
        part_n = basis_n * rescaling
        part_d = -basis_d * expected_values
        design_matrix = np.concatenate([part_n, part_d]).T
        pars = np.linalg.lstsq(design_matrix, expected_values, rcond=-1)[0]

        # Compute fitted model amplitudes.
        npar_n = len(self.numer_degrees)
        pars_n = pars[:npar_n]
        pars_d = pars[npar_n:]
        amplitudes_model = np.dot(pars_n, basis_n) / (1 + np.dot(pars_d, basis_d))

        # Convert parameters to the original scale.
        pars = np.zeros(self.npar)
        pars[:npar_n] = pars_n / self.scales["freq_scale"] ** self.numer_degrees
        pars[npar_n:] = pars_d / self.scales["freq_scale"] ** self.denom_degrees

        return pars, amplitudes_model

    def compute(
        self, freqs: NDArray[float], pars: NDArray[float], *, deriv: int = 0
    ) -> list[NDArray[float]]:
        """See :py:meth:`SpectrumModel.compute`."""
        if not isinstance(deriv, int):
            raise TypeError("Argument deriv must be integer.")
        if deriv < 0:
            raise ValueError("Argument deriv must be zero or positive.")
        if freqs.ndim != 1:
            raise TypeError("Argument freqs must be a 1D array.")
        npar_n = len(self.numer_degrees)
        npar_d = len(self.denom_degrees)
        pars = np.asarray(pars)
        if pars.shape[-1] != npar_n + npar_d:
            raise ValueError("The number of parameters does not match the model.")

        # Construct two bases of monomials.
        basis_n = np.power.outer(freqs, self.numer_degrees).T
        basis_d = np.power.outer(freqs, self.denom_degrees).T
        pars_n = pars[..., :npar_n]
        pars_d = pars[..., npar_n:]

        # Compute model amplitudes and derivatives.
        num = np.einsum("...p,pf->...f", pars_n, basis_n)
        denom = 1 + np.einsum("...p,pf->...f", pars_d, basis_d)
        results = [num / denom]
        vec_shape = pars.shape[:-1]
        if deriv >= 1:
            model_grad = np.empty((*vec_shape, npar_n + npar_d, *freqs.shape))
            np.einsum("pf,...f->...pf", basis_n, 1 / denom, out=model_grad[..., :npar_n, :])
            block_n = model_grad[..., :npar_n, :]
            np.einsum(
                "pf,...f->...pf", basis_d, -results[0] / denom, out=model_grad[..., npar_n:, :]
            )
            results.append(model_grad)
        if deriv >= 2:
            model_hess = np.zeros((*vec_shape, npar_n + npar_d, npar_n + npar_d, *freqs.shape))
            np.einsum(
                "...pf,...f,qf->...pqf",
                block_n,
                1 / denom,
                -basis_d,
                out=model_hess[..., :npar_n, npar_n:, :],
            )
            np.einsum(
                "...pqf->...qpf",
                model_hess[..., :npar_n, npar_n:, :],
                out=model_hess[..., npar_n:, :npar_n, :],
            )
            np.einsum(
                "...f,pf,qf->...pqf",
                2 * results[0] / denom**2,
                basis_d,
                basis_d,
                out=model_hess[..., npar_n:, npar_n:, :],
            )
            results.append(model_hess)
        if deriv >= 3:
            raise ValueError("Third or higher derivatives are not supported.")
        return results

    def derive_props(self, props: dict[str, NDArray[float]]):
        """Add the autocorrelation integral (and other properties) derived from the parameters."""
        acint_props = {
            "acint": props["pars"][0],
            "acint_var": props["pars_covar"][0, 0],
        }
        props.update(acint_props)


@attrs.define
class LorentzModel(PadeModel):
    """A model for the spectrum with a Lorentzian peak at zero frequency plus some white noise.

    This is a special case of the PadeModel with
    `numer_degrees` = [0, 2] and `denom_degrees` = [2].
    Furthermore, it will only accept parameters that correspond
    to a well-defined exponential correlation time.

    For too small cutoffs (covering only the peak of the Lorentzian and not its decay),
    the estimates of the Lorentzian width, and consequently the exponential correlation time,
    become statistically unreliable.
    In this regime, the assumption of maximum a posteriori probability (MAP),
    on which STACIE relies to fit the model and estimate parameter uncertainties, also breaks down.
    Unreliable MAP estimates are inferred from
    the relative error of the exponential correlation time
    divided by the relative error of the autocorrelation integral.
    This implementation uses the ratio in two ways:

    1. When the ratio exceeds a predefined threshold (default value 100),
       the cutoff criterion is set to infinity.
    2. If the ratio remains below this threshold,
       the ratio times a weight (default value 1.0) is added to the cutoff criterion.

    Note that this is an empirical penalty to mitigate MAP-related issues.
    Because the penalty is expressed as a ratio of relative errors, it is dimensionless
    and insensitive to the overall uncertainty of the spectrum.

    The hyperparameters `ratio_weight` and `ratio_threshold` may be tuned
    to adjust the sensitivity of the heuristic, but it is recommended to keep
    their default values.
    """

    ratio_weight: float = attrs.field(default=1.0, kw_only=True)
    """The penalty for the cutoff criterion is this weight times the ratio of relative errors."""

    ratio_threshold: float = attrs.field(default=100.0, kw_only=True)
    """A threshold for the ratio of relative errors used to set the cutoff criterion to Inf."""

    # Hard-code the polynomial degrees of the Pade model.
    numer_degrees: NDArray[int] = attrs.field(init=False, factory=lambda: np.array([0, 2]))
    denom_degrees: NDArray[int] = attrs.field(init=False, factory=lambda: np.array([2]))

    @property
    def name(self):
        return "lorentz()"

    def derive_props(self, props: dict[str, NDArray[float]]):
        """Add the autocorrelation integral (and other properties) derived from the parameters.

        The exponential correlation time is derived from the parameters,
        if the fitted model has a maximum at zero frequency.
        If not, the "criterion" is set to infinity and the "msg" is set accordingly,
        to discard the current fit from the average over the cutoff frequencies.
        """
        super().derive_props(props)
        # Try to deduce the exponential correlation time from the parameters.
        pars = props["pars"]
        if (
            # real correlation time
            pars[2] > 0
            # positive lorentzian amplitude, maximum in the origin
            and pars[0] * pars[2] > pars[1]
        ):
            # The following estimates of the exponential correlation time and its variance
            # are only valid for small variances.
            covar = props["pars_covar"]
            tau = np.sqrt(pars[2]) / (2 * np.pi)
            dtau_dp = 1 / (4 * np.pi * np.sqrt(pars[2]))
            tau_var = covar[2, 2] * dtau_dp**2
            tau_props = {
                "corrtime_exp": tau,
                "corrtime_exp_var": tau_var,
                "exp_block_time": tau * np.pi / 10,
                "exp_block_time_var": tau_var * np.pi / 10,
                "exp_simulation_time": 20 * tau * np.pi,
                "exp_simulation_time_var": 20 * tau_var * np.pi,
            }
            props.update(tau_props)
            if "criterion" in props:
                # Empirical penalty to eliminate or down-weight cutoffs
                # for which the maximum a posteriori approximation is expected to break down.
                relerr_corrtime = tau_var**0.5 / tau
                relerr_acint = props["acint_var"] ** 0.5 / props["acint"]
                if relerr_corrtime > self.ratio_threshold * relerr_acint:
                    props["criterion"] = np.inf
                    props["msg"] = (
                        f"rel.err. tau_exp > {self.ratio_threshold:.1e} x rel.err. ac integral"
                    )
                else:
                    ratio = relerr_corrtime / relerr_acint
                    props["criterion"] += self.ratio_weight * ratio

        else:
            # If we fail to find the exponential correlation time with a decent relative error,
            # we discard the entire estimate at this cutoff frequency.
            props["criterion"] = np.inf
            props["msg"] = "No correlation time estimate available."


def guess(
    freqs: NDArray[float],
    ndofs: NDArray[float],
    amplitudes: NDArray[float],
    weights: NDArray[float],
    model: SpectrumModel,
    rng: np.random.Generator,
    nonlinear_budget: int,
):
    """Guess initial values of the parameters for a model.

    Parameters
    ----------
    freqs
        The frequencies for which the model spectrum amplitudes are computed.
    ndofs
        The number of degrees of freedom at each frequency.
    amplitudes
        The amplitudes of the spectrum.
    weights
        Fitting weights, in range [0, 1], to use for each grid point.
    model
        The model for which the parameters are guessed.
    rng
        The random number generator.
    nonlinear_budget
        The number of samples of the nonlinear parameters is computed as
        ``nonlinear_budget ** num_nonlinear``, where ``num_nonlinear`` is the number
        of nonlinear parameters.

    Returns
    -------
    pars
        An initial guess of the parameters.
    """
    if not isinstance(nonlinear_budget, int) or nonlinear_budget < 1:
        raise ValueError("Argument nonlinear_budget must be a strictly positive integer.")

    # Get the mask for the nonlinear parameters
    nonlinear_mask = model.get_par_nonlinear()
    num_nonlinear_pars = nonlinear_mask.sum()

    # If there are no nonlinear parameters, we can directly guess the linear parameters.
    if num_nonlinear_pars == 0:
        return _guess_linear(freqs, ndofs, amplitudes, weights, model, [], nonlinear_mask)[1]

    # Otherwise, we need to sample the nonlinear parameters and guess the linear parameters.
    nonlinear_samples = model.sample_nonlinear_pars(rng, nonlinear_budget**num_nonlinear_pars)
    best = None
    for nonlinear_pars in nonlinear_samples:
        cost, pars = _guess_linear(
            freqs, ndofs, amplitudes, weights, model, nonlinear_pars, nonlinear_mask
        )
        if best is None or best[0] > cost:
            best = cost, pars
    return best[1]


def _guess_linear(
    freqs: NDArray[float],
    ndofs: NDArray[float],
    amplitudes: NDArray[float],
    weights: NDArray[float],
    model: SpectrumModel,
    nonlinear_pars: NDArray[float],
    nonlinear_mask: NDArray[bool],
) -> tuple[float, NDArray[float]]:
    """Guess initial values of the linear parameters for a model.

    Parameters
    ----------
    freqs
        The frequencies for which the model spectrum amplitudes are computed.
    ndofs
        The number of degrees of freedom at each frequency.
    amplitudes
        The amplitudes of the spectrum.
    weights
        Fitting weights, in range [0, 1], to use for each grid point.
    model
        The model for which the parameters are guessed.
    nonlinear_pars
        The values of the nonlinear parameters.
    nonlinear_mask
        A boolean mask for the nonlinear parameters.

    Returns
    -------
    cost
        The cost of the guess.
    pars
        An initial guess of the parameters.
    """
    # Perform a weighted least squares fit to guess the linear parameters.
    linear_pars, amplitudes_model = model.solve_linear(
        freqs, ndofs, amplitudes, weights, nonlinear_pars
    )

    # Combine the linear and nonlinear parameters
    pars = np.zeros(model.npar)
    pars[nonlinear_mask] = nonlinear_pars
    pars[~nonlinear_mask] = linear_pars

    # Fix invalid guesses
    invalid_mask = model.which_invalid(pars)
    par_scales = model.par_scales
    pars[invalid_mask] = par_scales[invalid_mask]
    if not model.valid(pars):
        raise RuntimeError("Invalid guess could not be fixed. This should never happen.")

    # If the model has zero or negative values,
    # increase the value of the constant term to fix the guess.
    def negerr(x):
        """Returns a negative value if the spectrum is less than 1e-3 of ``amp_scale``.

        Parameters
        ----------
        x
            A correction to the first element of the parameter vector.
        """
        mod_pars = pars.copy()
        mod_pars[0] += x
        return model.compute(freqs, mod_pars)[0].min() - model.scales["amp_scale"] * 1e-3

    if negerr(0) < 0.0:
        bracket = _find_bracket(negerr, 0, par_scales[0])
        xopt = brentq(negerr, bracket[0], bracket[1])
        pars[0] += xopt
        amplitudes_model = model.compute(freqs, pars)[0]

    # Compute the cost
    delta = amplitudes - amplitudes_model
    cost = np.einsum("i,i,i", delta, delta, weights)
    return cost, pars


def _find_bracket(f, x0, x1):
    """Fit a bracket with a sign change.

    This function assumes that a sign change is present either in [x0, x1]
    or outside the bracket on the side of x1.
    """
    f0 = f(x0)
    f1 = f(x1)
    for _ in range(100):
        if np.sign(f0) != np.sign(f1):
            break
        x2 = x1 + (x1 - x0) * 2
        f2 = f(x2)
        x0, x1 = x1, x2
        f0, f1 = f1, f2
    else:
        raise RuntimeError("Failed to find a bracket with a sign change.")
    return x0, x1



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Utility to prepare the spectrum and other inputs for given sequences."""

from collections.abc import Iterable
from itertools import repeat
from typing import Self

import attrs
import numpy as np
from numpy.typing import ArrayLike, NDArray

__all__ = ("Spectrum", "compute_spectrum")


@attrs.define
class Spectrum:
    """Container class holding all the inputs for the autocorrelation integral estimate."""

    mean: float = attrs.field(converter=float)
    """The mean of the input sequences multiplied by the square root of the prefactor."""

    variance: float = attrs.field(converter=float)
    """The variance of the input sequences multiplied by the prefactor."""

    timestep: float = attrs.field(converter=float)
    """The time between two subsequent elements in the given sequence."""

    nstep: int = attrs.field(converter=int)
    """The number of time steps in the input."""

    freqs: NDArray[float] = attrs.field()
    """The equidistant frequency axis of the spectrum."""

    ndofs: NDArray[float] = attrs.field()
    """The number of independent contributions to each amplitude."""

    amplitudes: NDArray[float] = attrs.field()
    """The spectrum amplitudes averaged over the given input sequences."""

    amplitudes_ref: NDArray[float] | None = attrs.field(default=None)
    """Optionally, the known analytical model of the power spectrum, on the same frequency grid."""

    @property
    def nfreq(self) -> int:
        """The number of RFFT frequency grid points."""
        return len(self.freqs)

    def without_zero_freq(self) -> Self:
        """Return a copy without the DC component."""
        if self.freqs[0] != 0.0:
            raise ValueError("The zero frequency has already been removed.")
        variance = self.variance - self.mean**2
        nindep = self.ndofs[0]
        variance *= (nindep * self.nstep) / (nindep * self.nstep - 1)
        return attrs.evolve(
            self,
            variance=variance,
            freqs=self.freqs[1:],
            ndofs=self.ndofs[1:],
            amplitudes=self.amplitudes[1:],
            amplitudes_ref=None if self.amplitudes_ref is None else self.amplitudes_ref[1:],
        )


def compute_spectrum(
    sequences: Iterable[NDArray[float]] | NDArray[float],
    *,
    prefactors: Iterable[NDArray[float]] | NDArray[float] | None = 1.0,
    timestep: float = 1,
    include_zero_freq: bool = True,
) -> Spectrum:
    r"""Compute a spectrum and store all inputs for ``estimate_acint`` in a ``Spectrum`` instance.

    The spectrum amplitudes are computed as follows:

    .. math::

        C_k = \frac{1}{M}\sum_{m=1}^M \frac{F_m h}{2 N} \left|
            \sum_{n=0}^{N-1} x^{(m)}_n \exp\left(-i \frac{2 \pi n k}{N}\right)
        \right|^2

    where:

    - :math:`F_m` is the given prefactor (may be different for each sequence),
    - :math:`h` is the timestep,
    - :math:`N` is the number of time steps in the input sequences,
    - :math:`M` is the number of independent sequences,
    - :math:`x^{(m)}_n` is the value of the :math:`m`-th sequence at time step :math:`n`,
    - :math:`k` is the frequency index.

    The sum over :math:`m` simply averages spectra obtained from different sequences.
    The factor :math:`F_m h/ 2 N` normalizes the spectrum so that its zero-frequency limit
    is an estimate of the autocorrelation integral.

    Parameters
    ----------
    sequences
        The input sequences, which can have several forms.
        If ``prefactors`` is not ``None``, it can be:

        - An array with shape ``(nindep, nstep)`` or ``(nstep,)``.
          In case of a 2D array, each row is a time-dependent sequence.
          In case of a 1D array, a single sequence is used.
        - An iterable whose items are arrays as described in the previous point.
          This option is convenient when a single array does not fit in memory.

        If ``prefactors`` is ``None``:

        - A tuple of a prefactor (or an array of prefactors) and a sequences array,
          either 1D or 2D as described above.
        - An iterable whose items are tuples of a prefactor (or an array of prefactors)
          and a sequences array, either 1D or 2D as described above.

        All sequences are assumed to be statistically independent and have length ``nstep``.
        (Time correlations within one sequence are fine, obviously.)
        We recommend using multiple independent sequences to reduce uncertainties.
        Arrays must be used. (lists of floating point values are not supported.)
    prefactors
        A positive factor to be multiplied with the autocorrelation function
        to give it a physically meaningful unit.
        This argument can be given in multiple forms:

        - None, in which case the sequences are assumed to be
          one or more (prefactors, sequences) tuples.
        - A single floating point value: the same prefactor is used for all input sequences.
        - A single array with shape ``(nindep,)``:
          each sequence is multiplied with the corresponding prefactor.
        - An iterable whose items are of the form described in the previous two points.
          In this case, the sequences must also be given as an iterable with the same length.
    timestep
        The time step of the input sequence.
    include_zero_freq
        When set to False, the DC component of the spectrum is discarded.

    Returns
    -------
    spectrum
        A ``Spectrum`` object holding all the inputs needed to estimate
        the integral of the autocorrelation function.
    """
    # Handle tuple (prefactor, sequences) case
    if isinstance(sequences, tuple):
        prefactors, sequences = sequences

    # Handle single-array case, assume consistency between sequences and prefactor.
    if isinstance(sequences, np.ndarray):
        sequences = [sequences]
        prefactors = [prefactors]

    # Process iterable of arrays
    if isinstance(sequences, Iterable):
        nindep = 0
        nstep = None
        amplitudes = 0
        total = 0
        total_sq = 0
        if prefactors is None:
            iterator = sequences
        elif isinstance(prefactors, Iterable):
            iterator = zip(prefactors, sequences, strict=True)
        else:
            iterator = zip(repeat(prefactors), sequences, strict=False)
        for item_prefactors, item_sequences in iterator:
            nstep, item_nindep, item_amplitudes, item_total, item_total_sq = _process_sequences(
                item_sequences, item_prefactors, nstep
            )
            nindep += item_nindep
            amplitudes += item_amplitudes
            total += item_total
            total_sq += item_total_sq
    else:
        raise TypeError("The sequence argument must be an array or an iterable of arrays.")

    # Frequency axis and scale of amplitudes
    freqs = np.fft.rfftfreq(nstep, d=timestep)
    amplitudes *= timestep / nindep

    # Number of "degrees of freedom" (contributions) to each amplitude
    ndofs = np.full(freqs.shape, 2 * nindep, dtype=float)
    ndofs[0] = nindep
    if len(freqs) % 2 == 0:
        ndofs[-1] = nindep

    # Remove DC component, useful for inputs that oscillate about a non-zero average.
    # The variance is calculated consistently:
    # - If the DC component is removed, the variance is calculated with respect to the mean.
    # - Otherwise, the variance is calculated with respect to zero.
    mean = total / nindep
    if include_zero_freq:
        variance = total_sq / nindep
    else:
        ndofs = ndofs[1:]
        freqs = freqs[1:]
        amplitudes = amplitudes[1:]
        variance = total_sq / nindep - mean**2
        variance *= (nstep * nindep) / (nstep * nindep - 1)

    return Spectrum(mean, variance, timestep, nstep, freqs, ndofs, amplitudes)


def _process_sequences(
    sequences: ArrayLike, prefactors: ArrayLike, nstep: int | None
) -> tuple[int, NDArray[float], float, float]:
    """Process a batch of sequences and compute the spectrum.

    Parameters
    ----------
    sequences
        The input sequences, which can be in two forms:
        This is an array with shape ``(nindep, nstep)`` or ``(nstep,)``.
        In case of a 2D array, each row is a time-dependent sequence.
        In case of a 1D array, a single sequence is used.
    prefactors
        A factor to be multiplied with the autocorrelation function
        to give it a physically meaningful unit.
        If multiple prefactors are given, each one is multiplied with the corresponding sequence.
    nstep
        The expected number of time steps in the input sequences or None.
        This is only used for consistency checking.

    Returns
    -------
    nstep
        The number of time steps in the input.
    nindep
        The number of independent sequences.
    amplitudes
        The spectrum amplitudes averaged over the given input sequences.
    total
        The sum of the input sequences.
    total_sq
        The sum of the squares of the input sequences.
    """
    sequences = np.asarray(sequences)
    prefactors = np.asarray(prefactors)

    # Handle single sequence case
    if sequences.ndim == 1:
        sequences = sequences.reshape(1, -1)
    elif sequences.ndim != 2:
        raise ValueError("Sequences must be a 1D or 2D array.")

    if prefactors.ndim == 0:
        prefactors = np.full(sequences.shape[0], prefactors)

    # Get basic parameters of the input sequences.
    if nstep is None:
        nstep = sequences.shape[1]
    elif nstep != sequences.shape[1]:
        raise ValueError("All sequences must have the same length.")
    nindep = sequences.shape[0]

    # Multiply the square root of (prefactor / 2) with the sequences.
    sequences = np.sqrt(prefactors)[:, None] * sequences

    # Compute the spectrum.
    # We already divide by nstep here to keep the order of magnitude under control.
    amplitudes = 0.5 * (abs(np.fft.rfft(sequences, axis=1)) ** 2).sum(axis=0) / nstep

    # Compute the variance of the input sequences.
    total = sequences.sum() / nstep
    total_sq = np.linalg.norm(sequences) ** 2 / nstep

    return nstep, nindep, amplitudes, total, total_sq



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Test configuration."""

import matplotlib as mpl
import numdifftools as nd
import numpy as np
import pytest

__all__ = ("check_curv", "check_deriv", "check_gradient", "check_hessian")

mpl.rcParams["backend"] = "Agg"
mpl.rcParams["font.size"] = 8.0
mpl.rcParams["figure.constrained_layout.use"] = True
mpl.rcParams["axes.spines.top"] = False
mpl.rcParams["axes.spines.right"] = False


def check_deriv(func, x0):
    """Check the derivative with STACIE's convention for returning gradients."""
    deriv = func(x0, deriv=1)[1]
    num_deriv, info = nd.Derivative(lambda x: func(x)[0], full_output=True)(x0)
    error = np.clip(info.error_estimate, 1e-15, np.inf)
    assert deriv / error == pytest.approx(num_deriv / error, abs=100)


def check_curv(func, x0):
    """Check the curvature with STACIE's convention for returning gradients."""
    curv = func(x0, deriv=2)[2]
    num_curv, info = nd.Derivative(lambda x: func(x, deriv=1)[1], full_output=True)(x0)
    error = np.clip(info.error_estimate, 1e-15, np.inf)
    assert curv / error == pytest.approx(num_curv / error, abs=100)


def check_gradient(func, x0):
    """Check the gradient with STACIE's convention for returning gradients."""
    grad = func(x0, deriv=1)[1]
    num_grad, info = nd.Gradient(lambda x: func(x)[0], full_output=True)(x0)
    error = np.clip(info.error_estimate, 1e-15, np.inf)
    if num_grad.ndim == 2:
        num_grad = num_grad.T
        error = error.T
    assert grad / error == pytest.approx(num_grad / error, abs=100)


def check_hessian(func, x0):
    """Check the Hessian with STACIE's convention for returning gradients."""
    hess = func(x0, deriv=2)[2]
    num_hess, info = nd.Gradient(lambda x: func(x, deriv=1)[1], full_output=True)(x0)
    error = np.clip(info.error_estimate, 1e-15, np.inf)
    assert hess / error == pytest.approx(num_hess / error, abs=100)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Cost function to optimize models for the low-frequency part of the spectrum."""

import attrs
import numpy as np
from numpy.typing import NDArray
from scipy.special import digamma, gammaln, polygamma

from .model import SpectrumModel

__all__ = ("LowFreqCost", "entropy_gamma", "logpdf_gamma")


@attrs.define
class LowFreqCost:
    """Cost function to fit a model to the low-frequency part of the spectrum."""

    freqs: NDArray[float] = attrs.field()
    """The frequencies for which the spectrum amplitudes are computed."""

    ndofs: NDArray[int] = attrs.field()
    """The number of independent contributions to each spectrum amplitude."""

    amplitudes: NDArray[float] = attrs.field()
    """The actual spectrum amplitudes at frequencies in ``self.freqs``."""

    weights: NDArray[float] = attrs.field()
    """The fitting weights for each grid point."""

    model: SpectrumModel = attrs.field()
    """The model to be fitted to the spectrum."""

    def __call__(self, pars: NDArray[float], *, deriv: int = 0) -> float:
        """Evaluate the cost function and its derivatives.

        Parameters
        ----------
        pars
            The parameter vector for which the loss function must be computed.
        deriv
            The order of derivatives of the cost function to include.

        Returns
        -------
        results
            A list with the cost function and the requested derivatives.
        """
        # Prepare result arrays, with inf and nan by default.
        # These will be overwritten if the model is valid.
        pars = np.asarray(pars)
        vec_shape = pars.shape[:-1]
        par_shape = pars.shape[-1:]
        results = [np.full(vec_shape, np.inf)]
        if deriv >= 1:
            results.append(np.full(vec_shape + par_shape, np.nan))
        if deriv >= 2:
            results.append(np.full(vec_shape + par_shape + par_shape, np.nan))
        if deriv >= 3:
            raise ValueError("Third or higher derivatives are not supported.")

        mask = self.model.valid(pars)
        if not mask.any():
            return results

        # Compute the model spectrum and its derivatives.
        amplitudes_model = self.model.compute(
            self.freqs, pars if mask.ndim == 0 else pars[mask], deriv=deriv
        )
        alphas = 0.5 * self.ndofs

        # Only continue with parameters for which the model does not become negative.
        # Small positive values are also excluded to avoid underflows.
        pos_mask = (amplitudes_model[0] > 1e-30).all(axis=-1)
        if not pos_mask.any():
            return results
        if mask.ndim > 0:
            amplitudes_model = [am[pos_mask] for am in amplitudes_model]
            mask[mask] = pos_mask
        del pos_mask

        # Log-likelihood computed with the scaled Chi-squared distribution.
        # The Gamma distribution is used because the scale parameter is easily incorporated.
        thetas = [am / alphas for am in amplitudes_model]
        ll_terms = logpdf_gamma(self.amplitudes, alphas, thetas[0], deriv=deriv)
        nlp = self.model.neglog_prior(pars[mask] if mask.ndim > 0 else pars, deriv=deriv)
        results[0][mask] = -np.einsum("...i,i->...", ll_terms[0], self.weights) + nlp[0]
        if deriv >= 1:
            results[1][mask] = (
                -np.einsum("...pi,...i,i->...p", thetas[1], ll_terms[1], self.weights) + nlp[1]
            )
        if deriv >= 2:
            results[2] = (
                -(
                    np.einsum(
                        "...pi,...qi,...i,i->...pq", thetas[1], thetas[1], ll_terms[2], self.weights
                    )
                    + np.einsum("...pqi,...i,i->...pq", thetas[2], ll_terms[1], self.weights)
                )
                + nlp[2]
            )
        return results

    def expected(self, pars: NDArray[float]) -> NDArray[float]:
        """Compute the expected value and variance of the cost function.

        Parameters
        ----------
        pars
            The model parameters.
            Vectorization is not supported yet.

        Returns
        -------
        expected, variance
            The expected value and variance of the cost function.
        """
        pars = np.asarray(pars)
        amplitudes_model = self.model.compute(self.freqs, pars)
        alphas = 0.5 * self.ndofs
        thetas = amplitudes_model[0] / alphas
        return (
            np.dot(entropy_gamma(alphas, thetas)[0], self.weights),
            np.dot(varlogp_gamma(alphas), self.weights**2),
        )


def logpdf_gamma(
    x: NDArray[float], alpha: NDArray[float], theta: NDArray[float], *, deriv: int = 0
):
    """Compute the logarithm of the probability density function of the Gamma distribution.

    Parameters
    ----------
    x
        The argument of the PDF (random variable).
        Array with shape ``(nfreq,)``.
    alpha
        The shape parameter.
        Array with shape ``(nfreq,)``.
    theta
        The scale parameter.
        Array with shape ``(..., nfreq,)``.
    deriv
        The order of the derivatives toward theta to compute: 0, 1 or 2.

    Returns
    -------
    results
        A list of results (function value and requested derivatives.)
        All elements have the same shape as the ``theta`` array.
    """
    alpha = np.asarray(alpha)
    theta = np.asarray(theta)
    ratio = np.asarray(x) / theta
    results = [-gammaln(alpha) - np.log(theta) + (alpha - 1) * np.log(ratio) - ratio]
    if deriv >= 1:
        results.append((ratio - alpha) / theta)
    if deriv >= 2:
        results.append((alpha - 2 * ratio) / theta**2)
    if deriv >= 3:
        raise ValueError("Third or higher derivatives are not supported.")
    return results


def entropy_gamma(alpha: NDArray[float], theta: NDArray[float], *, deriv: int = 0):
    """Compute the entropy of the Gamma distribution.

    Parameters
    ----------
    alpha
        The shape parameter.
    theta
        The scale parameter.
    deriv
        The order of the derivatives toward theta to compute: 0, 1 or 2.

    Returns
    -------
    results
        A list of results (function value and requested derivatives.)
        All elements have the same shape as the ``alpha`` and ``theta`` arrays.
    """
    alpha = np.asarray(alpha)
    theta = np.asarray(theta)
    results = [alpha + np.log(theta) + gammaln(alpha) + (1 - alpha) * digamma(alpha)]
    if deriv >= 1:
        results.append(1 / theta)
    if deriv >= 2:
        results.append(-1 / theta**2)
    if deriv >= 3:
        raise ValueError("Third or higher derivatives are not supported.")
    return results


def varlogp_gamma(alpha: NDArray[float]) -> NDArray[float]:
    """Compute the variance of the log-probability density function of the Gamma distribution.

    Parameters
    ----------
    alpha
        The shape parameter.

    Returns
    -------
    var
        The variance of the log-probability density function.
        Array with shape ``(alpha,)``.
    """
    alpha = np.asarray(alpha)
    return (alpha - 1) ** 2 * polygamma(1, alpha) - alpha + 2



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Tests for ``stacie.utils``."""

import numpy as np
import pytest
from numpy.testing import assert_equal

from stacie.utils import block_average, mixture_stats, robust_dot, robust_posinv, split


def test_split_sequences():
    assert_equal(split([1, 2, 3, 4, 5, 6], 2), [[1, 2, 3], [4, 5, 6]])
    assert_equal(split([1, 2, 3, 4, 5, 6, 7], 2), [[1, 2, 3], [4, 5, 6]])
    assert_equal(split([[1, 2, 3, 4, 5, 6]], 2), [[1, 2, 3], [4, 5, 6]])
    assert_equal(split([[1, 2, 3, 4, 5, 6, 7]], 2), [[1, 2, 3], [4, 5, 6]])
    assert_equal(split([[1, 2, 3, 4], [5, 6, 7, 8]], 2), [[1, 2], [3, 4], [5, 6], [7, 8]])
    assert_equal(split([[1, 2, 3, 4, -1], [5, 6, 7, 8, -2]], 2), [[1, 2], [3, 4], [5, 6], [7, 8]])


def test_block_average():
    assert_equal(block_average([1, 2, 3, 4, 5, 6], 2), [1.5, 3.5, 5.5])
    assert_equal(block_average([1, 2, 3, 4, 5, 6, 7], 2), [1.5, 3.5, 5.5])
    assert_equal(block_average([[1, 2, 3, 4, 5, 6]], 2), [[1.5, 3.5, 5.5]])
    assert_equal(block_average([[1, 2, 3, 4, 5, 6, 7]], 2), [[1.5, 3.5, 5.5]])
    assert_equal(block_average([[1, 2, 3, 4], [5, 6, 7, 8]], 3), [[2.0], [6.0]])
    assert_equal(
        block_average(
            [
                [1, 2, 3, 4, 5, 6, 7, 8],
                [9, 10, 11, 12, 13, 14, 15, 16],
            ],
            4,
        ),
        [[2.5, 6.5], [10.5, 14.5]],
    )
    assert_equal(
        block_average([[[1, 2, 3, 4], [5, 6, 7, 8]], [[7, 8, 9, 10], [11, 12, 13, 14]]], 2),
        [[[1.5, 3.5], [5.5, 7.5]], [[7.5, 9.5], [11.5, 13.5]]],
    )


@pytest.mark.parametrize(
    "a",
    [
        [[8, 2], [2, 4]],
        [[7, 2, 3], [2, 6, 1], [3, 1, 9]],
    ],
)
def test_robust_posinv_simple(a):
    scales, evals, evecs, inv = robust_posinv(a)
    assert scales.shape == (len(a),)
    assert evals.shape == (len(a),)
    assert evecs.shape == (len(a), len(a))
    assert inv.shape == (len(a), len(a))
    assert (evals > 0).all()
    assert inv == pytest.approx(inv.T)
    assert np.dot(a, inv) == pytest.approx(np.eye(len(a)), abs=1e-10)


def test_robust_posinv_hard():
    basis = np.array(
        [
            [10, 1e15, -3e-15],
            [2, 1e18, -5e-14],
            [3.7, -5e17, 1e-18],
        ]
    )
    a = np.dot(basis.T, basis)
    scales, evals, evecs, inv = robust_posinv(a)
    assert scales.shape == (len(a),)
    assert evals.shape == (len(a),)
    assert evecs.shape == (len(a), len(a))
    assert inv.shape == (len(a), len(a))
    assert (evals > 0).all()
    print(inv)
    assert inv == pytest.approx(inv.T)
    half = np.dot(basis, inv)
    assert np.dot(half.T, half) == pytest.approx(inv, rel=1e-10)


@pytest.mark.parametrize("shape", [(3,), (3, 5)])
def test_robust_dot_simple(shape):
    rng = np.random.default_rng(0)
    basis = rng.standard_normal((3, 3))
    a = np.dot(basis.T, basis)
    scales, evals, evecs, inv = robust_posinv(a)
    other = rng.standard_normal(shape)
    result = robust_dot(scales, evals, evecs, other)
    assert result.shape == shape
    assert result == pytest.approx(np.dot(a, other))
    result = robust_dot(1 / scales, 1 / evals, evecs, other)
    assert result.shape == shape
    assert result == pytest.approx(np.dot(inv, other))


def test_mixture_stats():
    means = np.array([[1, 2], [3, 4]])
    covars = np.array([[[1.2, 2.5], [3.1, 1.1]], [[7.3, 8.1], [2.2, 3.1]]])
    weights = np.array([2.0, 8.0])
    mean1, covar1 = mixture_stats(means, covars, weights)
    assert mean1 == pytest.approx([2.6, 3.6])
    assert covar1 == pytest.approx(
        0.2 * covars[0]
        + 0.8 * covars[1]
        + 0.2 * np.outer(means[0] - mean1, means[0] - mean1)
        + 0.8 * np.outer(means[1] - mean1, means[1] - mean1)
    )
    mean2, covar2 = mixture_stats(means, np.einsum("ijj->ij", covars), weights)
    assert mean2 == pytest.approx(mean1)
    assert covar2 == pytest.approx(np.diag(covar1))



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Tests for ``stacie.conditioning``."""

import numpy as np
import pytest
from conftest import check_gradient, check_hessian

from stacie.conditioning import ConditionedCost


def function(x, deriv: int = 0):
    """Compute the product of all items in x, and its gradient and Hessian.

    Parameters
    ----------
    x
        The input vector.
        For vectorized calculations, use N-dimensional inputs of which the last axis
        corresponds to the parameter index.
    deriv
        The order of the derivative to compute. Default is 0.
    This is just a simple function to test conditioning implementation.
    """
    results = [np.prod(x, axis=-1)]
    if deriv >= 1:
        results.append(np.einsum("...,...i->...i", results[0], 1 / x))
    if deriv >= 2:
        hess = np.einsum("...i,...j->...ij", results[1], 1 / x)
        for i in range(x.shape[-1]):
            hess[..., i, i] = 0
        results.append(hess)
    return results


def test_vectorize_function():
    """Check that the function can be vectorized."""
    x0 = np.array([[1.0, 2.0, 4.0], [1.0, 3.0, 4.0]])
    results = function(x0, deriv=2)
    assert results[0].shape == (2,)
    assert results[1].shape == (2, 3)
    assert results[2].shape == (2, 3, 3)
    for i, one_x0 in enumerate(x0):
        one_results = function(one_x0, deriv=2)
        assert one_results[0].shape == ()
        assert one_results[1].shape == (3,)
        assert one_results[2].shape == (3, 3)
        assert np.all(results[0][i] == one_results[0])
        assert np.all(results[1][i] == one_results[1])
        assert np.all(results[2][i] == one_results[2])


def test_function_deriv1():
    x0 = np.array([1.0, 2.0, 3.0, 4.0])
    check_gradient(function, x0)


def test_function_deriv2():
    x0 = np.array([1.0, 2.0, 3.0, 4.0])
    check_hessian(function, x0)


def test_vectorize_conditioned_cost():
    """Check that the conditioned cost can be vectorized."""
    par_scales = np.array([5.0, 2.3, 7.0])
    cost = ConditionedCost(function, par_scales, 5.0)
    x0 = np.array([[1.0, 2.0, 3.0], [1.0, 3.0, 4.0]])
    results = cost(x0, deriv=2)
    assert results[0].shape == (2,)
    assert results[1].shape == (2, 3)
    assert results[2].shape == (2, 3, 3)
    for i, one_x0 in enumerate(x0):
        one_results = cost(one_x0, deriv=2)
        assert one_results[0].shape == ()
        assert one_results[1].shape == (3,)
        assert one_results[2].shape == (3, 3)
        assert np.all(results[0][i] == one_results[0])
        assert np.all(results[1][i] == one_results[1])
        assert np.all(results[2][i] == one_results[2])


def test_conditioned_cost():
    par_scales = np.array([1.0, 2.0, 3.0, 4.0])
    cost = ConditionedCost(function, par_scales, 5.0)
    x0 = np.array([0.1, 0.2, 0.3, 0.4])
    assert cost(x0, deriv=0) == pytest.approx([np.prod(x0 * par_scales) / 5.0])
    assert cost.from_reduced(x0) == pytest.approx(x0 * par_scales)
    assert cost.to_reduced(x0) == pytest.approx(x0 / par_scales)
    assert cost.funcgrad(x0)[0] == pytest.approx(cost(x0)[0])
    assert cost.funcgrad(x0)[1] == pytest.approx(cost(x0, deriv=1)[1])
    assert cost.hess(x0) == pytest.approx(cost(x0, deriv=2)[2])


def test_conditioned_cost_deriv1():
    par_scales = np.array([1.0, 2.0, 3.0, 4.0])
    cost = ConditionedCost(function, par_scales, 5.0)
    x0 = np.array([0.1, 0.2, 0.3, 0.4])
    check_gradient(cost, x0)


def test_conditioned_cost_deriv2():
    par_scales = np.array([1.0, 2.0, 3.0, 4.0])
    cost = ConditionedCost(function, par_scales, 5.0)
    x0 = np.array([0.1, 0.2, 0.3, 0.4])
    check_hessian(cost, x0)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Unit tests for ``stacie.synthetic``"""

import numpy as np
import pytest

from stacie.spectrum import compute_spectrum
from stacie.synthetic import generate


def test_shape_none():
    psd = np.array([1, 2, 3, 4, 5])
    nseq = 3
    sequences = generate(psd, 1.0, nseq)
    assert sequences.shape == (nseq, 8)


def test_shape_shorter():
    psd = np.array([1, 2, 3, 4, 5])
    nseq = 3
    sequences = generate(psd, 1.0, nseq, 6)
    assert sequences.shape == (nseq, 6)


def test_shape_longer():
    psd = np.array([1, 2, 3, 4, 5])
    nseq = 3
    with pytest.raises(ValueError):
        generate(psd, 1.0, nseq, 20)


@pytest.mark.parametrize("truncate", [False, True])
def test_psd(truncate: bool):
    # Set up a test case with a Lorentzian power spectral density.
    tau = 5.0
    h = 0.1
    nstep = 4096
    tmax = h * nstep
    nfreq = nstep // 2 + 1
    freq0 = 1 / tmax
    omegas = np.arange(nfreq) * 2 * np.pi * freq0
    psd = tau / (1 + (omegas * tau) ** 2)
    nseq = 1024
    if truncate:
        nstep = 2048
    rng = np.random.default_rng(42)
    sequences = generate(psd, h, nseq, nstep, rng)

    # Test the shape of the output.
    assert sequences.shape == (nseq, nstep)
    # Test the consistency with compute_spectrum.
    spectrum = compute_spectrum(sequences, prefactors=2.0, timestep=h)
    if truncate:
        # The spectrum subsampled due to the truncation.
        psd = psd[::2]
    assert spectrum.amplitudes[:512] == pytest.approx(psd[:512], rel=0.2)
    # Test the Plancherel theorem. Origins of the factors 2:
    # - One factor two from the RFFT. (The PSD is one-sided.)
    # - If truncated, there is a second factor 2, because the power spectrum
    #   in compute_spectrum has a factor 1/N, and the generate function
    #   is compatible with this convention by including factor 1/sqrt(N) in the inverse RFFT.
    true_variance = 2 * psd.sum() * freq0
    if truncate:
        true_variance *= 2
    assert (sequences**2).mean() == pytest.approx(true_variance, rel=0.1)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Cost function pre-conditioning."""

from collections.abc import Callable

import attrs
import numpy as np
from numpy.typing import NDArray

__all__ = ("ConditionedCost",)


@attrs.define
class ConditionedCost:
    """A wrapper for the cost function taking care of pre-conditioning.

    The goal of the pre-conditioner is to let the optimizer work with normalized parameters,
    and to scale the cost function to a normalized range, such that all quantities are close to 1,
    even if the spectra and the frequencies have very different orders of magnitude.
    """

    cost: Callable[[NDArray[float], int], list[NDArray[float]]] = attrs.field()
    par_scales: NDArray[float] = attrs.field()
    cost_scale: float = attrs.field()

    def __call__(self, pars: NDArray[float], *, deriv: int = 0) -> list[NDArray[float]]:
        """Evaluate the pre-conditioned cost function.

        Parameters
        ----------
        pars
            The parameters to evaluate the cost function at, in the original space.
            For vectorized calculations, use N-dimensional inputs of which the last axis
            corresponds to the parameters.
        deriv
            The order of the derivative to compute.

        Returns
        -------
        results
            The cost function value and its derivatives.
            In vectorized calculations, the last axis of the gradient
            and the last two of the Hessian correspond to the parameters.
        """
        if not isinstance(deriv, int):
            raise TypeError("Argument deriv must be integer.")
        if deriv < 0:
            raise ValueError("Argument deriv must be zero or positive.")
        pars_orig = pars * self.par_scales
        results_orig = self.cost(pars_orig, deriv=deriv)
        results = [results_orig[0] / self.cost_scale]
        if deriv == 0:
            return results
        results.append(results_orig[1] * (self.par_scales / self.cost_scale))
        if deriv == 1:
            return results
        results.append(
            results_orig[2] * (np.outer(self.par_scales, self.par_scales) / self.cost_scale)
        )
        if deriv == 2:
            return results
        raise NotImplementedError("Third and higher-order derivatives are not supported.")

    def to_reduced(self, pars: NDArray[float]) -> NDArray[float]:
        """Convert parameters from the original to the reduced space.

        Parameters
        ----------
        pars
            The parameters to convert, in the original space.

        Returns
        -------
        pars_reduced
            The parameters in the reduced space.
        """
        return pars / self.par_scales

    def from_reduced(self, pars: NDArray[float]) -> NDArray[float]:
        """Convert parameters from the reduced to the original space.

        Parameters
        ----------
        pars
            The parameters to convert, in the reduced space.

        Returns
        -------
        pars_orig
            The parameters in the original space.
        """
        return pars * self.par_scales

    def funcgrad(self, pars: NDArray[float]) -> tuple[float, NDArray[float]]:
        """Compute the cost function and the gradient.

        Parameters
        ----------
        pars
            The parameters, in the reduced space.

        Returns
        -------
        cost_reduced
            The cost normalized function value.
        """
        return self(pars, deriv=1)

    def hess(self, pars: NDArray[float]) -> NDArray[float]:
        """Compute the Hessian matrix of the cost function."""
        return self(pars, deriv=2)[2]



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Generate synthetic time-correlated data for algorithmic testing and validation."""

import numpy as np
from numpy.typing import NDArray


def generate(
    psd: NDArray[float],
    timestep: float,
    nseq: int,
    nstep: int | None = None,
    rng: np.random.Generator | None = None,
) -> NDArray[float]:
    """Generate sequences with a given power spectral density.

    Parameters
    ----------
    psd
        The power spectral density.
        The normalization of the PSD is consistent ``compute_spectrum`` when
        using ``prefactors=2.0`` and the given ``timestep`` as arguments.
        The empirical amplitudes of the spectrum will then be consistent with given PSD.
        Hence ``psd[0]`` is the ground truth of the autocorrelation integral.
    timestep
        The time between two subsequent elements in the sequence.
    nseq
        The number of sequences to generate.
    nstep
        The number of time steps in each sequence.
        When not given, the number of steps is ``2 * (len(psd) - 1)``.
        This argument can be used to truncate the sequences,
        which can be useful for creating aperiodic signals.
    rng
        The random number generator.

    Returns
    -------
    sequences
        The generated sequences, a 2D array with shape ``(nseq, nstep)``,
        where ``nstep = 2 * (len(psd) - 1)`` if not provided.
    """
    # Set default arguments
    nstep_max = 2 * (len(psd) - 1)
    if nstep is None:
        nstep = nstep_max
    elif nstep > nstep_max:
        raise ValueError(f"nstep must be at most {nstep_max}")
    if rng is None:
        rng = np.random.default_rng()

    # Generate sequences in the frequency domain and transform to the time domain.
    nfreq = len(psd)
    ft = rng.normal(0, 1, (nseq, nfreq)) + 1.0j * rng.normal(0, 1, (nseq, nfreq))
    ft[:, 0].imag = 0
    ft[:, -1].imag = 0
    ft[:, 1:-1] /= np.sqrt(2)
    ft *= np.sqrt(psd)
    return np.fft.irfft(ft)[:, :nstep] * np.sqrt(nstep_max / timestep)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Plot various aspects of the results of the autocorrelation integral estimate."""

import re

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from numpy.typing import NDArray
from scipy import stats

from .estimate import Result
from .spectrum import Spectrum
from .utils import UnitConfig, label_unit

__all__ = ("plot_results",)


def fixformat(s: str) -> str:
    """Replace standard scientific notation with prettier unicode formatting."""

    def repl(match):
        factor = match.group(1)
        exp = str(int(match.group(2)))
        if exp == "0":
            return factor
        return f"${factor}\\times 10^{{{exp}}}$"

    result = re.sub(r"\b([0-9.]+)e([0-9+-]+)\b", repl, s)
    result = re.sub(r"\binf\b", "∞", result)
    return re.sub(r"\bnan\b", "?", result)


def plot_results(
    path_pdf: str,
    rs: Result | list[Result],
    uc: UnitConfig | None = None,
    *,
    figsize: tuple = (7.5, 4.21875),
    legend: bool = True,
):
    """Generate a multi-page PDF with plots of the autocorrelation integral estimation.

    Parameters
    ----------
    path_pdf
        The PDF file where all the figures are stored.
    rs
        A single ``Result`` instance or a list of them.
        If the (first) result instance has ``spectrum.amplitudes_ref`` set,
        theoretical expectations are included.
        When multiple results instances are given,
        only the first one is plotted in blue.
        All remaining ones are plotted in light grey.
    uc
        The configuration of the units used for plotting.
    figsize
        The figure size tuple for matplotlib
    """
    # Prepare results list
    if isinstance(rs, Result):
        rs = [rs]

    # Prepare units
    if uc is None:
        uc = UnitConfig()

    with PdfPages(path_pdf) as pdf:
        for r in rs:
            fig, ax = plt.subplots(figsize=figsize)
            plot_fitted_spectrum(ax, uc, r, legend=legend)
            pdf.savefig(fig)
            plt.close(fig)

            if len(r.history) > 1:
                fig, axs = plt.subplots(2, 2, figsize=figsize)
                plot_extras(axs, uc, r)
                pdf.savefig(fig)
                plt.close(fig)

        if len(rs) > 1:
            if rs[0].spectrum.amplitudes_ref is not None:
                fig, ax = plt.subplots(figsize=figsize)
                plot_qq(ax, uc, rs)
                pdf.savefig(fig)
                plt.close(fig)

            fig, ax = plt.subplots(figsize=figsize)
            plot_acint_estimates(ax, uc, rs)
            pdf.savefig(fig)
            plt.close(fig)


REF_PROPS = {"ls": "--", "color": "k", "alpha": 0.5}


def plot_spectrum(ax: mpl.axes.Axes, uc: UnitConfig, s: Spectrum, nplot: int | None = None):
    """Plot the empirical spectrum."""
    if nplot is None or nplot > s.nfreq:
        nplot = s.nfreq
    ax.plot(
        s.freqs[:nplot] / uc.freq_unit,
        s.amplitudes[:nplot] / uc.acint_unit,
        "C0.",
        mew=0,
    )
    _plot_ref_spectrum(ax, uc, s, nplot)
    fmax = s.freqs[:nplot].max() / uc.freq_unit
    ax.set_xlim(-0.01 * fmax, fmax * 1.01)
    ax.set_xlabel(label_unit("Frequency", uc.freq_unit_str))
    ax.set_ylabel(label_unit("Amplitude", uc.acint_unit_str))
    ax.set_title("Spectrum")


def _plot_ref_spectrum(ax: mpl.axes.Axes, uc: UnitConfig, s: Spectrum, nplot: int):
    """Plot the reference spectrum."""
    if s.amplitudes_ref is not None:
        ax.plot(
            s.freqs[:nplot] / uc.freq_unit,
            s.amplitudes_ref[:nplot] / uc.acint_unit,
            **REF_PROPS,
        )


FIT_LEFT_TITLE_TEMPLATE = (
    "Model {model} \n"
    "${uc.acint_symbol}$ = {acint:{uc.acint_fmt}} ± {acint_std:{uc.acint_fmt}}"
    "{acint_unit_str}"
)

FIT_RIGHT_TITLE_TEMPLATE = (
    "$\\tau_\\text{{int}}$ = {corrtime_int:{uc.time_fmt}}"
    " ± {corrtime_int_std:{uc.time_fmt}}"
    "{time_unit_str}"
)

FIT_RIGHT_TITLE_TEMPLATE_EXP = (
    "$\\tau_\\text{{exp}}$ = {corrtime_exp:{uc.time_fmt}}"
    " ± {corrtime_exp_std:{uc.time_fmt}}"
    "{time_unit_str}"
)


def plot_fitted_spectrum(ax: mpl.axes.Axes, uc: UnitConfig, r: Result, *, legend: bool = True):
    """Plot the fitted model spectrum."""
    freqs = r.spectrum.freqs[: r.ncut]
    # Show fitting weight on top of the spectrum
    ax2 = ax.twinx()
    ax2.plot(freqs / uc.freq_unit, r.props["weights"], color="C3", ls=":")
    ax2.set_ylabel("Fitting weight", color="C3")
    ax2.set_ylim(-0.02, 1.02)
    ax2.spines["right"].set_color("red")
    ax2.spines["right"].set_visible(True)
    ax2.tick_params(axis="y", colors="red")

    # The empirical spectrum.
    plot_spectrum(ax, uc, r.spectrum, r.ncut)

    # Model spectrum.
    neff = int(np.ceil(r.neff))
    freqs = r.spectrum.freqs[:neff]
    alphas = 0.5 * r.spectrum.ndofs[:neff]
    mean = r.props["amplitudes_model"][0][:neff]
    std_fit = np.sqrt(
        np.einsum(
            "ij,ik,jk->k",
            r.props["pars_covar"],
            r.props["amplitudes_model"][1][:, :neff],
            r.props["amplitudes_model"][1][:, :neff],
        )
    )
    ax.plot(freqs / uc.freq_unit, mean / uc.acint_unit, color="C2")
    ax.plot(
        freqs / uc.freq_unit,
        stats.norm.ppf(uc.clb, mean, std_fit) / uc.acint_unit,
        color="C2",
        ls="--",
    )
    ax.plot(
        freqs / uc.freq_unit,
        stats.norm.ppf(uc.cub, mean, std_fit) / uc.acint_unit,
        color="C2",
        ls="--",
        label=f"{uc.clevel:.0%} CI fitted model",
    )
    ax.fill_between(
        freqs / uc.freq_unit,
        stats.gamma.ppf(uc.clb, alphas, scale=mean / alphas) / uc.acint_unit,
        stats.gamma.ppf(uc.cub, alphas, scale=mean / alphas) / uc.acint_unit,
        color="C2",
        alpha=0.3,
        lw=0,
        label=f"{uc.clevel:.0%} CI sampling PSD",
    )
    ax.axvline(r.fcut / uc.freq_unit, ymax=0.1, color="k")
    ax.set_ylim(top=r.spectrum.amplitudes[: r.ncut].max() / uc.acint_unit * 1.1)
    if legend:
        ax.legend(loc="best")
    # Info in title
    fields = {
        "uc": uc,
        "model": r.model.name,
        "acint": r.acint / uc.acint_unit,
        "acint_std": r.acint_std / uc.acint_unit,
        "acint_unit_str": "" if uc.acint_unit_str == "1" else " " + uc.acint_unit_str,
        "corrtime_int": r.corrtime_int / uc.time_unit,
        "corrtime_int_std": r.corrtime_int_std / uc.time_unit,
        "time_unit_str": "" if uc.time_unit_str == "1" else " " + uc.time_unit_str,
    }
    ax.set_title("")
    ax.set_title(fixformat(FIT_LEFT_TITLE_TEMPLATE.format(**fields)), loc="left")
    if "corrtime_exp" in r.props:
        fields["corrtime_exp"] = r.corrtime_exp / uc.time_unit
        fields["corrtime_exp_std"] = r.corrtime_exp_std / uc.time_unit
        ax.set_title(
            fixformat(
                FIT_RIGHT_TITLE_TEMPLATE_EXP.format(**fields)
                + "\n"
                + FIT_RIGHT_TITLE_TEMPLATE.format(**fields)
            ),
            loc="right",
        )
    else:
        ax.set_title("\n" + fixformat(FIT_RIGHT_TITLE_TEMPLATE.format(**fields)), loc="right")


def plot_extras(axs: NDArray[mpl.axes.Axes], uc: UnitConfig, r: Result):
    plot_cutoff_weight(axs[0, 0], uc, r)
    plot_sanity(axs[1, 0], uc, r)
    axs[0, 0].sharex(axs[1, 0])
    axs[0, 0].set_xlabel(None)
    plot_uncertainty(axs[0, 1], uc, r)
    plot_evals(axs[1, 1], uc, r)
    axs[0, 1].sharex(axs[1, 1])
    axs[0, 1].set_xlabel(None)


def plot_cutoff_weight(ax: mpl.axes.Axes, uc: UnitConfig, r: Result):
    """Plot the cutoff criterion as a function of cutoff frequency."""
    fcuts = np.array([props["fcut"] for props in r.history])
    criteria = np.array([props["criterion"] for props in r.history])
    criteria -= criteria.min()
    probs = np.exp(-criteria)
    probs /= probs.sum()
    ax.plot(fcuts / uc.freq_unit, probs, color="C1")
    ax.set_xscale("log")
    ax.set_xlabel(label_unit("Cutoff frequency", uc.freq_unit_str))
    ax.set_ylabel(f"{r.cutoff_criterion.name} weight")


def plot_sanity(ax: mpl.axes.Axes, uc: UnitConfig, r: Result):
    fcuts = np.array([props["fcut"] for props in r.history])
    low = -1
    high = 4
    for key, color in ("cost", "C0"), ("criterion", "C2"):
        zscore = np.array([props[f"{key}_zscore"] for props in r.history])
        ax.plot(fcuts / uc.freq_unit, zscore, color=color, label=key.title())
        low = min(low, zscore.min())
        high = max(high, r.props[f"{key}_zscore"] * 1.2)
    ax.set_xscale("log")
    ax.set_xlabel(label_unit("Cutoff frequency", uc.freq_unit_str))
    ax.set_ylabel("Z-score")
    ax.axhline(0, color="k", ls="--")
    ax.axhline(2, color="k", ls="--")
    ax.set_ylim(low - 0.2, high)
    ax.legend()


def plot_uncertainty(ax: mpl.axes.Axes, uc: UnitConfig, r: Result):
    """Plot the autocorrelation integral and uncertainty as a function fo cutoff frequency."""
    fcuts = []
    acints = []
    acint_stds = []
    for props in r.history:
        fcuts.append(props["fcut"])
        acints.append(props["acint"])
        acint_stds.append(props["acint_std"])
    fcuts = np.array(fcuts)
    acints = np.array(acints)
    acint_stds = np.array(acint_stds)

    ax.plot(fcuts / uc.freq_unit, acints / uc.acint_unit, "C3")
    ax.fill_between(
        fcuts / uc.freq_unit,
        stats.norm.ppf(uc.clb, acints, acint_stds) / uc.acint_unit,
        stats.norm.ppf(uc.cub, acints, acint_stds) / uc.acint_unit,
        color="C3",
        alpha=0.3,
        lw=0,
    )
    fcut_weights = np.array([props["fcut_weight"] for props in r.history])
    fcut_weights /= fcut_weights.max()
    mask = fcut_weights > 0.01
    ax.scatter(
        fcuts[mask] / uc.freq_unit,
        acints[mask] / uc.acint_unit,
        s=30,
        c=fcut_weights[mask],
        marker="o",
        linewidth=0,
        cmap="Greys",
        vmin=0,
        vmax=1,
    )
    if r.spectrum.amplitudes_ref is not None:
        limit = r.spectrum.amplitudes_ref[0]
        ax.axhline(limit / uc.acint_unit, **REF_PROPS)
    ax.set_xscale("log")
    ax.set_xlabel(label_unit("Cutoff frequency", uc.freq_unit_str))
    ax.set_ylabel(label_unit(f"${uc.acint_symbol}$ ({uc.clevel:.0%} CI)", uc.acint_unit_str))


def plot_evals(ax: mpl.axes.Axes, uc: UnitConfig, r: Result):
    """Plot the eigenvalues of the Hessian as a function of the cutoff frequency."""
    fcuts = np.array([props["fcut"] for props in r.history])
    all_evals = np.array([props["cost_hess_rescaled_evals"] for props in r.history])
    fcut_weights = np.array([props["fcut_weight"] for props in r.history])
    fcut_weights /= fcut_weights.max()

    ax.plot(fcuts / uc.freq_unit, all_evals, color="C4")
    mask = fcut_weights > 0.01
    for evals in all_evals.T:
        ax.scatter(
            fcuts[mask] / uc.freq_unit,
            evals[mask],
            s=30,
            c=fcut_weights[mask],
            marker="o",
            linewidth=0,
            cmap="Greys",
            vmin=0,
            vmax=1,
        )
    ax.set_xscale("log")
    ax.set_xlabel(label_unit("Cutoff frequency", uc.freq_unit_str))
    ax.set_ylabel("Hessian Eigenvalues")
    ax.set_yscale("log")


def plot_all_models(ax: mpl.axes.Axes, uc: UnitConfig, r: Result):
    """Plot all fitted model spectra (for all tested cutoffs)."""
    fcut_weights = np.array([props["fcut_weight"] for props in r.history])
    fcut_weights /= fcut_weights.max()
    nplot = 0
    for i, props in enumerate(r.history):
        ncut = props["ncut"]
        nplot = max(nplot, ncut)
        mean = props["amplitudes_model"][0]
        freqs = r.spectrum.freqs[:ncut]
        if fcut_weights[i] > 0.01:
            plot_kwargs = {"alpha": fcut_weights[i], "zorder": 2.5, "color": "C2"}
        else:
            plot_kwargs = {"alpha": 0.1, "zorder": 0.5, "color": "k"}
        ax.plot(freqs / uc.freq_unit, mean / uc.acint_unit, **plot_kwargs)
    _plot_ref_spectrum(ax, uc, r.spectrum, nplot)
    # Print the number of fitted model spectra in the title to show how many models were tested.
    ax.set_title(f"Model spectra ({len(r.history)} fits)", wrap=True)
    ax.set_xlabel(label_unit("Frequency", uc.freq_unit_str))
    ax.set_ylabel(label_unit("Amplitude", uc.acint_unit_str))
    ax.set_ylim(bottom=0)


def plot_qq(ax: mpl.axes.Axes, uc: UnitConfig, rs: list[Result]):
    """Make a qq-plot between the predicted and expected distribution of AC integral estimates.

    This plot function assumes the true integral is known.
    """
    cdfs = (np.arange(len(rs)) + 0.5) / len(rs)
    quantiles = stats.norm.ppf(cdfs)
    limit = rs[0].spectrum.amplitudes_ref[0]
    normed_errors = np.array([(r.acint - limit) / r.acint_std for r in rs])
    normed_errors.sort()
    distance = abs(quantiles - normed_errors).mean()
    ax.scatter(quantiles, normed_errors, c="C0", s=3)
    ax.plot([-2, 2], [-2, 2], **REF_PROPS)
    ax.set_xlabel("Normal quantiles")
    ax.set_ylabel("Sorted normalized errors")
    ax.set_title(f"QQ Plot (Wasserstein Distance = {distance:.4f})")


RELERR_TEMPLATE = """\
MRE = {mre:.1f} %
RMSRE = {rmsre:.1f} %
RMSRF = {rmsrf:.1f} %
RMSPRE = {rmspre:.1f} %
"""


def rms(x):
    return np.sqrt((x**2).mean())


def plot_acint_estimates(ax: mpl.axes.Axes, uc: UnitConfig, rs: list[Result]):
    """Plot the sorted autocorrelation integral estimates and their uncertainties."""
    values = np.array([r.acint for r in rs])
    stds = np.array([r.acint_std for r in rs])
    order = values.argsort()
    values = values[order]
    stds = stds[order]
    ax.errorbar(
        np.arange(len(rs)),
        values / uc.acint_unit,
        [
            (values - stats.norm.ppf(uc.clb, values, stds)) / uc.acint_unit,
            (stats.norm.ppf(uc.cub, values, stds) - values) / uc.acint_unit,
        ],
        fmt="o",
        lw=1,
        ms=2,
        ls="none",
    )
    if rs[0].spectrum.amplitudes_ref is not None:
        limit = rs[0].spectrum.amplitudes_ref[0]
        ax.axhline(limit / uc.acint_unit, **REF_PROPS)
        relative_errors = 100 * (values - limit) / values
        mre = relative_errors.mean()
        ax.text(
            0.05,
            0.95,
            RELERR_TEMPLATE.format(
                mre=mre,
                rmsre=rms(relative_errors),
                rmsrf=rms(relative_errors - mre),
                rmspre=rms(stds / values) * 100,
            ),
            transform=ax.transAxes,
            ha="left",
            va="top",
            linespacing=1.5,
        )
    ax.set_xlabel("Rank")
    ax.set_ylabel(label_unit("Mean and uncertainty", uc.acint_unit_str))
    ax.set_title("Autocorrelation Integral")



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Tests for ``stacie.model``"""

import numpy as np
import pytest
from conftest import check_gradient, check_hessian

from stacie.model import ExpPolyModel, PadeModel, guess

NFREQ = 10
FREQS = np.linspace(0, 0.5, NFREQ)
AMPLITUDES_REF = np.linspace(2, 1, NFREQ)
WEIGHTS = 1 - FREQS**2
TIMESTEP = 1.2


def check_vectorize_compute(model, pars_ref, broadcast=False):
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    pars_ref = np.array(pars_ref)
    amplitudes = model.compute(FREQS, pars_ref, deriv=2)
    nvec, npar = pars_ref.shape
    assert amplitudes[0].shape == (nvec, len(FREQS))
    if broadcast:
        assert amplitudes[1].shape == (1, npar, len(FREQS))
        assert amplitudes[2].shape == (1, npar, npar, len(FREQS))
    else:
        assert amplitudes[1].shape == (nvec, npar, len(FREQS))
        assert amplitudes[2].shape == (nvec, npar, npar, len(FREQS))
    for i, one_pars_ref in enumerate(pars_ref):
        one_amplitudes = model.compute(FREQS, one_pars_ref, deriv=2)
        assert one_amplitudes[0].shape == (len(FREQS),)
        assert one_amplitudes[1].shape == (npar, len(FREQS))
        assert one_amplitudes[2].shape == (npar, npar, len(FREQS))
        assert (one_amplitudes[0] == amplitudes[0][i]).all()
        if broadcast:
            assert (one_amplitudes[1] == amplitudes[1][0]).all()
            assert (one_amplitudes[2] == amplitudes[2][0]).all()
        else:
            assert (one_amplitudes[1] == amplitudes[1][i]).all()
            assert (one_amplitudes[2] == amplitudes[2][i]).all()


def check_vectorize_prior(model, pars_ref):
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    pars_ref = np.array(pars_ref)
    prior = model.neglog_prior(pars_ref, deriv=2)
    nvec, npar = pars_ref.shape
    assert prior[0].shape == (nvec,)
    assert prior[1].shape == (nvec, npar)
    assert prior[2].shape == (nvec, npar, npar)
    for i, one_pars_ref in enumerate(pars_ref):
        one_prior = model.neglog_prior(one_pars_ref, deriv=2)
        assert one_prior[0].shape == ()
        assert one_prior[1].shape == (npar,)
        assert one_prior[2].shape == (npar, npar)
        assert (one_prior[0] == prior[0][i]).all()
        assert (one_prior[1] == prior[1][i]).all()
        assert (one_prior[2] == prior[2][i]).all()


def test_exppoly_npar():
    assert ExpPolyModel([0, 2, 4]).npar == 3


PARS_REF_POLY = [
    [-12.0, 3.4, 78.3],
    [9.0, 8.1],
    [-0.1, 0.02, -0.7, 0.3],
    [0.0, 0.0, 0.0],
    [0.0, 3.0],
]


@pytest.mark.parametrize("npar", [2, 3])
def test_vectorize_exppoly_compute(npar: int):
    pars_ref = [p for p in PARS_REF_POLY if len(p) == npar]
    check_vectorize_compute(ExpPolyModel(list(range(npar))), pars_ref)


@pytest.mark.parametrize("npar", [2, 3])
def test_vectorize_exppoly_prior(npar: int):
    pars_ref = [p for p in PARS_REF_POLY if len(p) == npar]
    check_vectorize_prior(ExpPolyModel(list(range(npar))), pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_POLY)
def test_gradient_exppoly_compute(pars_ref):
    pars_ref = np.array(pars_ref)
    model = ExpPolyModel(list(range(len(pars_ref))))
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_gradient(lambda pars, *, deriv=0: model.compute(FREQS, pars, deriv=deriv), pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_POLY)
def test_hessian_exppoly_compute(pars_ref):
    pars_ref = np.array(pars_ref)
    model = ExpPolyModel(list(range(len(pars_ref))))
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_hessian(lambda pars, *, deriv=0: model.compute(FREQS, pars, deriv=deriv), pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_POLY)
def test_gradient_exppoly_prior(pars_ref):
    pars_ref = np.array(pars_ref)
    model = ExpPolyModel(list(range(len(pars_ref))))
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_gradient(lambda pars, *, deriv=0: model.neglog_prior(pars, deriv=deriv), pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_POLY)
def test_hessian_exppoly_prior(pars_ref):
    pars_ref = np.array(pars_ref)
    model = ExpPolyModel(list(range(len(pars_ref))))
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_hessian(lambda pars, *, deriv=0: model.neglog_prior(pars, deriv=deriv), pars_ref)


def test_guess_exppoly():
    model = ExpPolyModel([0, 1, 2])
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    rng = np.random.default_rng(123)
    amplitudes = rng.normal(size=len(FREQS)) ** 2
    ndofs = np.full(len(FREQS), 2)
    pars_init = guess(FREQS, ndofs, amplitudes, WEIGHTS, model, rng, 10)
    assert len(pars_init) == 3
    assert np.isfinite(pars_init).all()


def test_pade_npar():
    assert PadeModel([0, 2], [2]).npar == 3


PARS_REF_PADE = [
    [0.0, 0.0, 0.0, 0.0],
    [0.5, -2.0, 0.4, -8.3],
    [1.3, 2.0, -0.4, 0.0],
    [0.0, 0.5, 3.2, 1.3],
    [0.2, 0.9, 0.1, 0.0],
]


@pytest.mark.parametrize("model", [PadeModel([0, 1, 2], [2]), PadeModel([0, 2], [1, 2])])
def test_vectorize_pade_compute(model):
    check_vectorize_compute(model, PARS_REF_PADE)


@pytest.mark.parametrize("model", [PadeModel([0, 1, 2], [2]), PadeModel([0, 2], [1, 2])])
def test_vectorize_pade_prior(model):
    check_vectorize_prior(model, PARS_REF_PADE)


@pytest.mark.parametrize("pars_ref", PARS_REF_PADE)
def test_gradient_pade(pars_ref):
    pars_ref = np.array(pars_ref)
    model = PadeModel([0, 1, 2], [2])
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_gradient(lambda pars, *, deriv=0: model.compute(FREQS, pars, deriv=deriv), pars_ref)


@pytest.mark.parametrize("pars_ref", PARS_REF_PADE)
def test_hessian_pade(pars_ref):
    pars_ref = np.array(pars_ref)
    model = PadeModel([0, 1, 2], [2])
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    check_hessian(lambda pars, *, deriv=0: model.compute(FREQS, pars, deriv=deriv), pars_ref)


def test_guess_pade():
    model = PadeModel([0, 2], [2])
    model.configure_scales(TIMESTEP, FREQS, AMPLITUDES_REF)
    rng = np.random.default_rng(123)
    amplitudes = rng.normal(size=len(FREQS)) ** 2
    ndofs = np.full(len(FREQS), 2)
    pars_init = guess(FREQS, ndofs, amplitudes, WEIGHTS, model, rng, 10)
    assert len(pars_init) == 3
    assert np.isfinite(pars_init).all()


def test_guess_pade_detailed():
    freqs = np.linspace(0, 1.0, NFREQ)
    model = PadeModel([0, 2], [2])
    model.configure_scales(TIMESTEP, freqs, AMPLITUDES_REF)
    pars_ref = np.array([3.0, 1.5, 2.0])
    amplitudes_ref = model.compute(freqs, pars_ref, deriv=0)[0]
    x = freqs / freqs[-1]
    assert amplitudes_ref == pytest.approx((3.0 + 1.5 * x**2) / (1.0 + 2.0 * x**2), rel=1e-10)
    ndofs = np.full(len(freqs), 20)
    pars_init_low, amplitudes_low = model.solve_linear(freqs, ndofs, amplitudes_ref, WEIGHTS, [])
    assert pars_init_low == pytest.approx(pars_ref, rel=1e-10)
    assert amplitudes_low == pytest.approx(amplitudes_ref, rel=1e-10)
    rng = np.random.default_rng(123)
    pars_init = guess(freqs, ndofs, amplitudes_ref, WEIGHTS, model, rng, 10)
    assert pars_ref == pytest.approx(pars_init, rel=1e-10)



# STACIE is a STable AutoCorrelation Integral Estimator.
# Copyright (C) 2024-2025 The contributors of the STACIE Python Package.
# See the CONTRIBUTORS.md file in the project root for a full list of contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# --
"""Unit tests for ``stacie.cutoff``."""

import numpy as np
import pytest

from stacie.cutoff import (
    CV2LCriterion,
    linear_weighted_regression,
)
from stacie.model import ExpPolyModel
from stacie.spectrum import compute_spectrum


def test_cv2l_preconditioned():
    nstep = 400
    rng = np.random.default_rng(42)
    spectrum = compute_spectrum(rng.standard_normal((10, nstep)))
    model = ExpPolyModel([0, 1])
    pars = np.array([0.1, 0.2])
    fcut = spectrum.freqs[nstep // 4]
    ncut = nstep // 3
    props = {
        "fcut": fcut,
        "ncut": ncut,
        "pars": pars,
        "switch_exponent": 20.0,
    }
    result1 = CV2LCriterion(regularize=False)(spectrum, model, props)
    result2 = CV2LCriterion(regularize=False, precondition=False)(spectrum, model, props)
    assert result1["criterion"] == pytest.approx(result2["criterion"], rel=1e-5)


def test_cv2l_regularize():
    nstep = 400
    rng = np.random.default_rng(42)
    spectrum = compute_spectrum(rng.standard_normal((10, nstep)))
    model = ExpPolyModel([0, 1])
    pars = np.array([0.1, 0.2])
    fcut = spectrum.freqs[nstep // 4]
    ncut = nstep // 3
    props = {
        "fcut": fcut,
        "ncut": ncut,
        "pars": pars,
        "switch_exponent": 20.0,
        "cost_hess_scales": np.array([1.0, 10.0]),
        "cost_hess_rescaled_evals": np.array([1.0, 2.0]),
        "cost_hess_rescaled_evecs": np.array([[1.0, 0.0], [0.0, 1.0]]),
    }
    result1 = CV2LCriterion()(spectrum, model, props)
    result2 = CV2LCriterion(precondition=False)(spectrum, model, props)
    assert result1["criterion"] == pytest.approx(result2["criterion"], rel=1e-5)


def test_linear_weighted_regression_lc():
    neq = 10
    npar = 4
    nw = 2
    rng = np.random.default_rng(42)
    dm = rng.standard_normal((neq, npar))
    ev = rng.standard_normal(neq)
    ws = rng.uniform(0, 1, (nw, neq))

    # Without linear combination
    xs1, cs1 = linear_weighted_regression(dm, ev, ws)
    assert xs1.shape == (nw, npar)
    assert cs1.shape == (nw, npar, nw, npar)
    assert cs1[0, :, 0] == pytest.approx(cs1[0, :, 0].T)
    assert cs1[1, :, 1] == pytest.approx(cs1[1, :, 1].T)
    assert cs1[1, :, 0] == pytest.approx(cs1[0, :, 1].T)
    xd = xs1[0] - xs1[1]
    cd = cs1[0, :, 0] + cs1[1, :, 1] - (cs1[0, :, 1] + cs1[1, :, 0])
    assert np.all(np.isfinite(xd))
    assert np.all(np.isfinite(cd))

    # With linear combination
    lc = np.array([[1.0, -1.0]])
    xs2, cs2 = linear_weighted_regression(dm, ev, ws, lc)
    assert xs2.shape == (1, npar)
    assert cs2.shape == (1, npar, 1, npar)
    assert xs2[0] == pytest.approx(xd)
    assert cs2[0, :, 0] == pytest.approx(cd)


def test_linear_weighted_regression_uncorrelated():
    neq = 10
    npar = 4
    nw = 2
    rng = np.random.default_rng(123)
    dm = rng.standard_normal((neq, npar))
    ev = rng.standard_normal(neq)
    ws = np.ones((nw, neq))
    ws[0, 5:] = 0
    ws[1, :5] = 0
    xs, cs = linear_weighted_regression(dm, ev, ws)
    assert xs[0] == pytest.approx(np.linalg.lstsq(dm[:5], ev[:5], rcond=None)[0], rel=1e-5)
    assert xs[1] == pytest.approx(np.linalg.lstsq(dm[5:], ev[5:], rcond=None)[0], rel=1e-5)
    assert cs[0, :, 1] == pytest.approx(np.zeros((npar, npar)), abs=1e-5)
    assert cs[1, :, 0] == pytest.approx(np.zeros((npar, npar)), abs=1e-5)


def test_linear_weighted_regression_example():
    neq = 50
    npar = 2
    nw = 2
    rng = np.random.default_rng(0)
    t = np.linspace(-1, 1, neq)
    dm = np.vstack((np.ones(neq), t)).T
    ref = 5 * t**2
    ws = np.array([(1 + t) / 2, (1 - t) / 2])

    all_xs = []
    cs_mean = 0
    nrep = 10000
    for _ in range(nrep):
        ev = rng.standard_normal(neq) + ref
        xs, cs = linear_weighted_regression(dm, ev, ws)
        all_xs.append(xs)
        cs_mean += cs
    all_xs = np.array(all_xs)
    cs_mean /= nrep
    assert all_xs.shape == (nrep, nw, npar)
    assert cs_mean[0, :, 0] == pytest.approx(np.cov(all_xs[:, 0, :], rowvar=False), abs=2e-3)
    assert cs_mean[1, :, 1] == pytest.approx(np.cov(all_xs[:, 1, :], rowvar=False), abs=2e-3)
    assert cs_mean.reshape((nw * npar, nw * npar)) == pytest.approx(
        np.cov(all_xs[:, 0, :], all_xs[:, 1, :], rowvar=False), abs=2e-3
    )
